#include "wins.h"
#include "dialogs.h"
#include "out.h"
#include "parser.h"

GRIDDIALOG* GridDialog (SREAL size, SINTEGER fill) {
 GRIDDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (GRIDDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Set Grid Parameters", GTK_WINDOW (GETMAINWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> sizeEntry = CreateEntry ("Grid Size", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 toRet -> fillEntry = CreateEntry ("Fill around", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 text = g_strdup_printf ("%.4f", size);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> sizeEntry), text);
 g_free (text);
 text = g_strdup_printf ("%ld", fill);
 gtk_entry_set_text (GTK_ENTRY (toRet -> fillEntry), text);
 g_free (text);
 return toRet;
}

void DestroyGridDialog (GRIDDIALOG* grid) {
 gtk_widget_destroy (grid -> dialog);
 g_slice_free1 (sizeof (GRIDDIALOG), grid);
}

COUNTERENTRYDIALOG* CounterEntryDialog (GtkWindow* parent, SCOUNTER count, gchar* name, gchar* caption) {
 COUNTERENTRYDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (COUNTERENTRYDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons (name, GTK_WINDOW (parent),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> entry = CreateEntry (caption, GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 text = g_strdup_printf ("%ld", count);
 gtk_entry_set_text (GTK_ENTRY (toRet -> entry), text);
 g_free (text);

 return toRet;
}

void DestroyCounterEntryDialog (COUNTERENTRYDIALOG* toDel) {
 gtk_widget_destroy (toDel -> dialog);
 g_slice_free1 (sizeof (COUNTERENTRYDIALOG), toDel);
}

CENTREDIALOG* CentreDialog () {
 CENTREDIALOG* toRet;
 GtkWidget* separator;

 toRet = g_slice_alloc (sizeof (CENTREDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Centre map", GTK_WINDOW (GETMAPWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
 toRet -> latEntry = CreateEntry ("Latitude", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> lonEntry = CreateEntry ("Longitude", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 gtk_entry_set_text (GTK_ENTRY (toRet -> latEntry), "0.0000");
 gtk_entry_set_text (GTK_ENTRY (toRet -> lonEntry), "0.0000");
 return toRet;
}

void DestroyCentreDialog (CENTREDIALOG* centre) {
 gtk_widget_destroy (centre -> dialog);
 g_slice_free1 (sizeof (CENTREDIALOG), centre);
}

MAPLIMITSDIALOG* MapLimitsDialog (SREAL n, SREAL s, SREAL e, SREAL w) {
 MAPLIMITSDIALOG* toRet;
 GtkWidget* separator;
 gchar* text;

 toRet = g_slice_alloc (sizeof (MAPLIMITSDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Set map limits", GTK_WINDOW (GETMAPWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> northEntry = CreateEntry ("North (Latitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> southEntry = CreateEntry ("South (Latitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 toRet -> eastEntry = CreateEntry ("East (Longitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> westEntry = CreateEntry ("West (Longitude)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 text = g_strdup_printf ("%.4f", n);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> northEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", s);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> southEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", e);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> eastEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", w);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (toRet -> westEntry), text);
 g_free (text);
 return toRet;
}

void DestroyMapLimitsDialog (MAPLIMITSDIALOG* mapLimits) {
 gtk_widget_destroy (mapLimits -> dialog);
 g_slice_free1 (sizeof (CENTREDIALOG), mapLimits);
}

SEARCHDIALOG* SearchDialog () {
 SEARCHDIALOG* toRet;
 GtkWidget* separator;

 toRet = g_slice_alloc (sizeof (SEARCHDIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Search", GTK_WINDOW (GETMAINWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> iterEntry = CreateEntry ("Iterations", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> equalEntry = CreateEntry ("Accept equals (percent)", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));
 toRet -> pageCheck = gtk_check_button_new_with_mnemonic ("_Page\'s heuristic");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> pageCheck, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> pageCheck);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);
 toRet -> keepCheck = gtk_check_button_new_with_mnemonic ("_Keep actual reconstructions");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> keepCheck, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> keepCheck);
 toRet -> allCheck = gtk_check_button_new_with_mnemonic ("Keep _all reconstructions");
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), toRet -> allCheck, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> allCheck);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);
 gtk_entry_set_text (GTK_ENTRY (toRet -> iterEntry), "100");
 gtk_entry_set_text (GTK_ENTRY (toRet -> equalEntry), "50");
 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toRet -> pageCheck), TRUE);
 return toRet;
}

void DestroySearchDialog (SEARCHDIALOG* searchDialog) {
 gtk_widget_destroy (searchDialog -> dialog);
 g_slice_free1 (sizeof (SEARCHDIALOG), searchDialog);
}

static gint CanWindowClose (GtkWidget* widget, gpointer data) {
 SFLAG* progressUp;

 progressUp = (SFLAG*) data;
 return *progressUp;
}

static void OnStopSearch (GtkButton* stopButton, gpointer data) {
 PROGRESSSTRUCT* pBar;

 pBar = (PROGRESSSTRUCT*) data;
 pBar -> isActive = FALSE;
}

SEARCHPROGDLG* SearchProgress (PROGRESSSTRUCT* progress, SFLAG mode) {
 SEARCHPROGDLG* toRet;
 GtkWidget* vBox;
 GtkWidget* label;

 toRet = g_slice_alloc (sizeof (SEARCHPROGDLG));
 toRet -> dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 toRet -> progressUp = TRUE;
 gtk_window_set_title (GTK_WINDOW (toRet -> dialog), "Searching...");
 g_signal_connect (G_OBJECT (toRet -> dialog), "delete_event", G_CALLBACK (CanWindowClose), &toRet -> progressUp);
 gtk_container_border_width (GTK_CONTAINER (toRet -> dialog), 0);
 gtk_window_set_modal (GTK_WINDOW (toRet -> dialog), TRUE);

 vBox = gtk_vbox_new (FALSE, 0);
 gtk_container_add (GTK_CONTAINER (toRet -> dialog), vBox);
 gtk_widget_show (vBox);
 label = gtk_label_new ("Trees");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 toRet -> treeProgress = gtk_progress_bar_new ();
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> treeProgress, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> treeProgress);
 label = gtk_label_new ("Iterations");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 toRet -> iterProgress = gtk_progress_bar_new ();
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> iterProgress, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> iterProgress);
 toRet -> stopButton = gtk_button_new_from_stock (GTK_STOCK_STOP);
 g_signal_connect (G_OBJECT (toRet -> stopButton), "clicked", G_CALLBACK (OnStopSearch), progress);
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> stopButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> stopButton);
 gtk_widget_set_sensitive (toRet -> stopButton, mode);

 toRet -> progressData = progress;
 gtk_widget_show (toRet -> dialog);
 return toRet;
}

gboolean UpdateSearchProgress (gpointer data) {
 SEARCHPROGDLG* prog;
 gchar* text;

 prog = (SEARCHPROGDLG*) data;
 if ((prog -> progressData -> isActive == FALSE) || (prog -> progressData -> actBar1 >= prog -> progressData -> numBar1)) {
   prog -> progressUp = FALSE;
   gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> treeProgress), 1.0);
   gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> iterProgress), 1.0);
   gtk_widget_destroy (prog -> dialog);
   g_slice_free1 (sizeof (SEARCHPROGDLG), prog);
   return FALSE;
 }
 text = g_strdup_printf ("Tree %ld of %ld", prog -> progressData -> actBar1, prog -> progressData -> numBar1);
 gtk_progress_bar_set_text (GTK_PROGRESS_BAR (prog -> treeProgress), text);
 g_free (text);
 gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> treeProgress),
                                            (gdouble) prog -> progressData -> actBar1 / prog -> progressData -> numBar1);
 text = g_strdup_printf ("Iteration %ld of %ld", prog -> progressData -> actBar2, prog -> progressData -> numBar2);
 gtk_progress_bar_set_text (GTK_PROGRESS_BAR (prog -> iterProgress), text);
 g_free (text);
 gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> iterProgress),
                                            (gdouble) prog -> progressData -> actBar2 / prog -> progressData -> numBar2);
 return TRUE;
}

void DestroyParsingProgressDialog (PARSINGPROGDLG* toDel) {
 gtk_widget_destroy (toDel -> dialog);
 g_slice_free1 (sizeof (PARSINGPROGDLG), toDel);
}

PARSINGPROGDLG* ParsingProgress (PROGRESSSTRUCT* progress) {
 PARSINGPROGDLG* toRet;
 GtkWidget* vBox;
 GtkWidget* label;

 toRet = g_slice_alloc (sizeof (PARSINGPROGDLG));
 toRet -> dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 toRet -> progressUp = TRUE;
 gtk_window_set_title (GTK_WINDOW (toRet -> dialog), "Parsing file...");
 g_signal_connect (G_OBJECT (toRet -> dialog), "delete_event", G_CALLBACK (CanWindowClose), &toRet -> progressUp);
 gtk_container_border_width (GTK_CONTAINER (toRet -> dialog), 0);
 gtk_window_set_modal (GTK_WINDOW (toRet -> dialog), TRUE);

 vBox = gtk_vbox_new (FALSE, 0);
 gtk_container_add (GTK_CONTAINER (toRet -> dialog), vBox);
 gtk_widget_show (vBox);
 label = gtk_label_new ("Parsing...");
 gtk_box_pack_start (GTK_BOX (vBox), label, TRUE, TRUE, 2);
 gtk_widget_show (label);
 toRet -> fileProgress = gtk_progress_bar_new ();
 gtk_box_pack_start (GTK_BOX (vBox), toRet -> fileProgress, TRUE, TRUE, 2);
 gtk_widget_show (toRet -> fileProgress);

 toRet -> progressData = progress;
 gtk_widget_show (toRet -> dialog);
 return toRet;
}

gboolean UpdateParsingProgress (gpointer data) {
 PARSINGPROGDLG* prog;

 prog = (PARSINGPROGDLG*) data;
 if ((prog -> progressData -> isActive == FALSE) || (prog -> progressData -> actBar1 >= prog -> progressData -> numBar1)) {
   prog -> progressUp = FALSE;
   gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> fileProgress), 1.0);
   gtk_widget_destroy (prog -> dialog);
   g_slice_free1 (sizeof (PARSINGPROGDLG), prog);
   return FALSE;
 }
 gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (prog -> fileProgress),
                                            (gdouble) prog -> progressData -> actBar1 / prog -> progressData -> numBar1);
 return TRUE;
}

static void PutGeo (GEOREFBROWSERDLG* dlg) {
 TERMINAL* term;
 gchar* text;
 GEOREF* geo;

 geo = dlg -> actGeo;
 text = g_strdup_printf ("%.4f", geo -> lat);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (dlg -> latEntry), text);
 g_free (text);
 text = g_strdup_printf ("%.4f", geo -> lon);
 CheckRealStr (text);
 gtk_entry_set_text (GTK_ENTRY (dlg -> lonEntry), text);
 g_free (text);
 term = (TERMINAL*) geo -> sp -> data;
 gtk_entry_set_text (GTK_ENTRY (dlg -> spEntry), term -> name);
 if (geo -> name != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> nameEntry), geo -> name);
 if (geo -> collec != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> collecEntry), geo -> collec);
 if (geo -> catalog != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> catalogEntry), geo -> catalog);
 if (geo -> url != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> urlEntry), geo -> url);
 if (geo -> reference != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> referenceEntry), geo -> reference);
 if (geo -> identifier != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> identifierEntry), geo -> identifier);
 if (geo -> sampleDate != NULL)
   gtk_entry_set_text (GTK_ENTRY (dlg -> sampleEntry), geo -> sampleDate);
 if (g_list_next (geo -> list) != NULL)
   gtk_widget_set_sensitive (dlg -> nextButton, TRUE);
 else
   gtk_widget_set_sensitive (dlg -> nextButton, FALSE);
 if (g_list_previous (geo -> list) != NULL)
   gtk_widget_set_sensitive (dlg -> prevButton, TRUE);
 else
   gtk_widget_set_sensitive (dlg -> prevButton, FALSE);
}

static void OnButtonPrevGeo (GtkButton* prevButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 GList* geoList;

 dlg = (GEOREFBROWSERDLG*) data;
 geoList = g_list_previous (dlg -> actGeo -> list);
 dlg -> actGeo = (GEOREF*) geoList -> data;
 PutGeo (dlg);
}

static void OnButtonNextGeo (GtkButton* nextButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 GList* geoList;

 dlg = (GEOREFBROWSERDLG*) data;
 geoList = g_list_next (dlg -> actGeo -> list);
 dlg -> actGeo = (GEOREF*) geoList -> data;
 PutGeo (dlg);
}

static void OnButtonApplyGeo (GtkButton* applyButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 SFLAG change;
 SREAL nuLat, nuLon;
 gchar* entry;
 GEOREF* geo;
 TERMINAL* actTerm;

 dlg = (GEOREFBROWSERDLG*) data;
 change = FALSE;
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> latEntry)));
 SetCommandLine (entry);
 g_free (entry);
 geo = dlg -> actGeo;
 if (GetCmdReal (&nuLat) != FALSE) {
   if ((nuLat > -90.0000) && (nuLat < 90.0000))
     change = TRUE;
   else
     nuLat = geo -> lat;
 }
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> lonEntry)));
 SetCommandLine (entry);
 g_free (entry);
 if (GetCmdReal (&nuLon) != FALSE) {
   if ((nuLon > -180.0000) && (nuLon < 180.0000))
     change = TRUE;
  else
     nuLon = geo -> lon;
 }
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> nameEntry)));
 if (g_strcmp0 (entry, geo -> name)) {
   actTerm = (TERMINAL*) geo -> sp -> data;
   geo -> name = AddSynonim (actTerm, entry);
 }
 g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> collecEntry)));
 if (g_strcmp0 (entry, geo -> collec)) {
   g_free (geo -> collec);
   geo -> collec = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> catalogEntry)));
 if (g_strcmp0 (entry, geo -> catalog)) {
   g_free (geo -> catalog);
   geo -> catalog = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> urlEntry)));
 if (g_strcmp0 (entry, geo -> url)) {
   g_free (geo -> url);
   geo -> url = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> referenceEntry)));
 if (g_strcmp0 (entry, geo -> reference)) {
   g_free (geo -> reference);
   geo -> reference = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> identifierEntry)));
 if (g_strcmp0 (entry, geo -> identifier)) {
   g_free (geo -> identifier);
   geo -> identifier = entry;
 }
 else
   g_free (entry);
 entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (dlg -> sampleEntry)));
 if (g_strcmp0 (entry, geo -> sampleDate)) {
   g_free (geo -> sampleDate);
   geo -> sampleDate = entry;
 }
 else
   g_free (entry);
 if (change == TRUE) {
   geo -> lat = nuLat;
   geo -> lon = nuLon;
   geo -> inclination = FROMLATTOINC (nuLat);
   geo -> azimuth = FROMLONTOAZM (nuLon);
   UpdateLimits (geo);
   PaintNewGeoRef (geo);
 }
 gtk_widget_destroy (dlg -> dialog);
 g_slice_free1 (sizeof (GEOREFBROWSERDLG), dlg);
}

static void OnButtonDelGeo (GtkButton* delButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;
 GEOREF* geo;

 dlg = (GEOREFBROWSERDLG*) data;
 geo = dlg -> actGeo;
 DeleteGeoRef (geo);
 ChangeGeo ();
 gtk_widget_destroy (dlg -> dialog);
 g_slice_free1 (sizeof (GEOREFBROWSERDLG), dlg);
}

static void OnButtonClose (GtkButton* closeButton, gpointer data) {
 GEOREFBROWSERDLG* dlg;

 dlg = (GEOREFBROWSERDLG*) data;
 gtk_widget_destroy (dlg -> dialog);
 g_slice_free1 (sizeof (GEOREFBROWSERDLG), dlg);
}

static gint OnCloseBrowser (GtkWidget* widget, gpointer data) {
 return TRUE;
}

GEOREFBROWSERDLG* GeoRefBrowser (GEOREF* geo, SFLAG mode) {
 GEOREFBROWSERDLG* toRet;
 GtkWidget* vBox;
 GtkWidget* hBox;
 GtkWidget* separator;
 GtkWidget* button;

 toRet = g_slice_alloc (sizeof (GEOREFBROWSERDLG));
 toRet -> dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 gtk_window_set_title (GTK_WINDOW (toRet -> dialog), "Record browser");
 g_signal_connect (G_OBJECT (toRet -> dialog), "delete_event", G_CALLBACK (OnCloseBrowser), NULL);
 gtk_container_border_width (GTK_CONTAINER (toRet -> dialog), 0);
 gtk_window_set_modal (GTK_WINDOW (toRet -> dialog), TRUE);
 vBox = gtk_vbox_new (FALSE, 0);
 gtk_container_add (GTK_CONTAINER (toRet -> dialog), vBox);
 gtk_widget_show (vBox);

 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 toRet -> prevButton = gtk_button_new_from_stock (GTK_STOCK_GO_BACK);
 g_signal_connect (G_OBJECT (toRet -> prevButton), "clicked", G_CALLBACK (OnButtonPrevGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> prevButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> prevButton);
 toRet -> nextButton = gtk_button_new_from_stock (GTK_STOCK_GO_FORWARD);
 g_signal_connect (G_OBJECT (toRet -> nextButton), "clicked", G_CALLBACK (OnButtonNextGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> nextButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> nextButton);

 toRet -> spEntry = CreateEntry ("Terminal", GTK_BOX (vBox));
 separator = gtk_hseparator_new ();
 toRet -> nameEntry = CreateEntry ("Record name", GTK_BOX (vBox));
 toRet -> latEntry = CreateEntry ("Latitude", GTK_BOX (vBox));
 toRet -> lonEntry = CreateEntry ("Longitude", GTK_BOX (vBox));
 separator = gtk_hseparator_new ();
 toRet -> collecEntry = CreateEntry ("Collection", GTK_BOX (vBox));
 toRet -> catalogEntry = CreateEntry ("Catalogue code", GTK_BOX (vBox));
 toRet -> urlEntry = CreateEntry ("Url", GTK_BOX (vBox));
 toRet -> referenceEntry = CreateEntry ("Reference", GTK_BOX (vBox));
 toRet -> identifierEntry = CreateEntry ("Identified by", GTK_BOX (vBox));
 toRet -> sampleEntry = CreateEntry ("Sample date", GTK_BOX (vBox));
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 hBox = gtk_hbox_new (TRUE, 2);
 gtk_box_pack_start (GTK_BOX (vBox), hBox, FALSE, TRUE, 2);
 gtk_widget_show (hBox);
 toRet -> applyButton = gtk_button_new_from_stock (GTK_STOCK_APPLY);
 g_signal_connect (G_OBJECT (toRet -> applyButton), "clicked", G_CALLBACK (OnButtonApplyGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> applyButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> applyButton);
 toRet -> delButton = gtk_button_new_from_stock (GTK_STOCK_DELETE);
 g_signal_connect (G_OBJECT (toRet -> delButton), "clicked", G_CALLBACK (OnButtonDelGeo), toRet);
 gtk_box_pack_start (GTK_BOX (hBox), toRet -> delButton, FALSE, TRUE, 2);
 gtk_widget_show (toRet -> delButton);
 gtk_widget_set_sensitive (toRet -> applyButton, mode);
 gtk_widget_set_sensitive (toRet -> delButton, mode);
 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vBox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 button = gtk_button_new_from_stock (GTK_STOCK_CLOSE);
 g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (OnButtonClose), toRet);
 gtk_box_pack_start (GTK_BOX (vBox), button, FALSE, TRUE, 2);
 gtk_widget_show (button);
 toRet -> actGeo = geo;

 gtk_editable_set_editable (GTK_EDITABLE (toRet -> spEntry), FALSE);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> nameEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> latEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> lonEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> collecEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> catalogEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> referenceEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> identifierEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> sampleEntry), mode);

 PutGeo (toRet);
 gtk_widget_show (toRet -> dialog);

 return toRet;
}

NODEDATADIALOG* NodeDataDialog (GNode* node, SFLAG mode) {
 NODEDATADIALOG* toRet;
 GtkWidget* separator;
 PHYLONODE* actNode;
 gchar* text;

 toRet = g_slice_alloc (sizeof (NODEDATADIALOG));
 toRet -> dialog = gtk_dialog_new_with_buttons ("Node", GTK_WINDOW (GETTREEWIN),
                       GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
                       GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);

 toRet -> nameEntry = CreateEntry ("Node name", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 toRet -> ageEntry = CreateEntry ("Node age", GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox));

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (toRet -> dialog) -> vbox), separator, TRUE, TRUE, 2);
 gtk_widget_show (separator);

 actNode = (PHYLONODE*) node -> data;
 if (actNode -> name != NULL)
   text = g_strdup_printf (actNode -> name);
 else
   text = g_strdup_printf ("unnamed");
 gtk_entry_set_text (GTK_ENTRY (toRet -> nameEntry), text);
 g_free (text);
  text = g_strdup_printf ("%ld", actNode -> age);
 gtk_entry_set_text (GTK_ENTRY (toRet -> ageEntry), text);
 g_free (text);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> nameEntry), mode);
 gtk_editable_set_editable (GTK_EDITABLE (toRet -> ageEntry), mode);
 return toRet;
}

void DestroyNodeDataDialog (NODEDATADIALOG* toDel) {
 gtk_widget_destroy (toDel -> dialog);
 g_slice_free1 (sizeof (NODEDATADIALOG), toDel);
}
