#ifndef GENERAL_HEADER_INCLUDED
#define GENERAL_HEADER_INCLUDED

/* ------ General header ------ */
/* The general header is based on glib, as the main project
   is under GTK+.
   But the general part, just uses the non GUI (i.e. glib)
   part of GTK+.
   */
#include <glib.h>

/* ------ Simple types ------ */
/* Synonims of the general types */

#define SINTEGER glong
#define SREAL gdouble
/* Use scounter to declare explicit unsigned values, it is negative for domain errors */
#define SCOUNTER SINTEGER

typedef guint16 SFLAG;      /* Control type */

/* Constant used to start strings */
#define VOIDSTRING "NoStr"

/* ------ Complex types ------ */

/* --- Arrays --- */
/* I use my own array function and structures because the strange behaviuor of the
   glib arrays. */

typedef struct tagINTARRAY INTARRAY;
struct tagINTARRAY {
 SINTEGER* data;
 SCOUNTER len;
 SCOUNTER max;
};

INTARRAY* CreateIntArray (SCOUNTER size);
#define AddToArray(x,y)         (x) -> data [(x) -> len ++] = (y)
#define InsertValToArray(x,y,i) (x) -> data [(i)] = (y)
#define GetFromArray(x,i)       (x) -> data [(i)]
#define GetArraySize(x)         (x) -> len
#define GetArrayMax(x)          (x) -> max
void DestroyIntArray (INTARRAY* array);

void CopyPtrArray (GPtrArray* dest, GPtrArray* src);
void ShufflePtrArray (GPtrArray* array);
void ShuffleSinglePtrArray (GPtrArray* array, SCOUNTER pos);

/* --- Functions for linked structures --- */
/* Some home-made functions for linked structures, as the standar implementation of glib
   states that does not guarantee to keep the original pointer on additions, removals,
   etc. on lists */

/* Insert an chain-element after prev */
GList* InsertListElement (GList* prev, gpointer data);
/* Deletes a list */
void DestroyList (GList* toDel, gsize sz);
void DestroyLinkElement (GList* toDel, gsize sz);
void DestroyLink (GList* toDel);
/* Put a list after last element of the destination list */
void PutAtListEnd (GList* dest, GList* toPut);
void AddChildren (GNode* toAdd, GNode* anc, GNode* lastSibling);
/* Get next valid node */
GNode* GetNextNode (GNode* actNode);
GNode* GetNextTerm (GNode* actNode);
GNode* GetPrevNode (GNode* actNode);
SFLAG IsDesc (GNode* probe, GNode* anc);

/* ------ Output log */
GString* outLog;

/* ------ Error interface ------ */
/* Used only for user errors (e.g. "bad file format") */
gchar* errorMessage;

/* This function must be defined in the user interface, and must include the RESETERROR macro */
#define RESETERROR g_free (errorMessage)
void OnError ();

#endif
