#include "wins.h"
#include "dialogs.h"
#include "parser.h"
#include "grid.h"
#include "voronoi.h"

typedef struct tagMAPWINDOWDATA MAPWINDATA;
struct tagMAPWINDOWDATA {
 GdkPixbuf* mapImg;
 GdkPixmap* pixMap;
 GtkWidget* drawable;

 GdkGC* blackPen;
 GdkGC* greenPen;
 GdkGC* bluePen;
 GdkGC* redPen;
 GdkGC* greyPen;
 GdkGC* whitePen;

 GtkWidget* menuDrawGrid;
 GtkWidget* menuDrawBarrier;
 GtkWidget* menuDrawRecords;
 GtkWidget* menuDrawCells;
 GtkWidget* menuDrawAll;
 GtkWidget* menuDrawRemoved;
 GtkWidget* menuBarrier;
 GtkWidget* menuBarrierVoronoi;
 GtkWidget* menuBarrierDelaunay;
 GtkWidget* menuBarrierMidTri;
 GtkWidget* menuBarrierAncestor;
 GtkWidget* menuBarrierCommon;

 GtkWidget* labelMousePos;

 gint clientX;
 gint clientY;
 SCOUNTER width;
 SCOUNTER height;
 SINTEGER originX;
 SINTEGER originY;
 SINTEGER startX;
 SINTEGER startY;
 SREAL north;
 SREAL south;
 SREAL west;
 SREAL east;
 gint prevX;
 gint prevY;
 gint radius;

 SREAL xScale;
 SREAL yScale;

 GNode* actNode;

 SFLAG mode;
 SFLAG barrierMode;
 SFLAG viewMode;
};

#define DRAW_MAP_MODE 0
#define EDIT_MAP_MODE 1

#define VIEW_RECORDS 0
#define VIEW_CELLS   1

#define DRAW_AS_VORONOI  0
#define DRAW_AS_DELAUNAY 1
#define DRAW_AS_MIDTRI   2

#define MENU_VORONOI  "voronoi_barrier"
#define MENU_DELAUNAY "delaunay_barrier"
#define MENU_MIDTRI "midtri_barrier"


static gint OnDelete (GtkWidget* mapWin, GdkEvent* event, gpointer data);
static gint OnExpose (GtkWidget* drawable, GdkEventExpose* event, gpointer data);
static gint OnConfigure (GtkWidget* drawable, GdkEvent* event, gpointer data);

static gint OnMotion (GtkWidget* drawable, GdkEventMotion* event, gpointer data);
static gint OnMouseButton (GtkWidget* drawable, GdkEventButton* event, gpointer data);

static void OnMenuLoad (GtkMenuItem* loadMenu, gpointer data);
static void OnMenuLimits (GtkMenuItem* limitsMenu, gpointer data);
static void OnMenuCentre (GtkMenuItem* centreMenu, gpointer data);
static void OnMenuRadius (GtkMenuItem* radiusMenu, gpointer data);
static void OnMenuViewMode (GtkMenuItem* drawCheckMenu, gpointer data);
static void OnMenuDrawCheck (GtkMenuItem* drawCheckMenu, gpointer data);
static void OnMenuBarrierMode (GtkMenuItem* modeMenu, gpointer data);
static void OnMenuBarrierOptions (GtkMenuItem* barrierMenu, gpointer data);

static void PrepareGeoRefs (SFLAG updateAll);
static void DrawMap ();
static void DrawDescs (GNode* refNode, GdkGC* pen);
static void DrawVicRec (GNode* refNode, GdkGC* pen, BITFIELD* outCells);
static void DrawDescsRec (GNode* refNode, GdkGC* pen);
static GEOREF* SearchNearestPoint (gint x, gint y);

static void DrawBarrier (GPtrArray* triangles);
static void DrawBarrierFromRec (RECONS* rec);
static void DrawBarrierFromAlloSet (PHYLONODE* node);

static MAPWINDATA mWinData;

void StartMapWindow () {
 GtkWidget* gBox;
 GtkWidget* menuBar;
 GtkWidget* menu;
 GtkAccelGroup* accelGroup;
 GSList* group;

 GdkColor blueColor = {0, 0x0000, 0x0000, 0xFFFF};
 GdkColor greenColor = {0, 0x0000, 0xFFFF, 0x0000};
 GdkColor redColor = {0, 0xFFFF, 0x0000, 0x0000};
 GdkColor greyColor = {0, 0x6000, 0x6000, 0x6000};

 mWinData . pixMap = NULL;
 mWinData . mapImg = NULL;
 mWinData . width = 720;
 mWinData . height = 360;
 mWinData . originX = mWinData . originY = 0;
 mWinData . startX = mWinData . startY = 0;
 mWinData . north = mWinData . west = 0;
 mWinData . south = 180.0000;
 mWinData . east = 360.0000;
 mWinData . radius = 1;

 mWinData . mode = DRAW_MAP_MODE;

 gBox = gtk_vbox_new (FALSE, 0);
 mapWindow = CreateWindow ("VIP: MapWindow", 100, 200, 800, 500, G_CALLBACK (OnDelete), gBox);

 /* --- Set menues --- */
 accelGroup = gtk_accel_group_new ();
 gtk_window_add_accel_group (GTK_WINDOW (mapWindow), accelGroup);
 menuBar = gtk_menu_bar_new ();
 gtk_box_pack_start (GTK_BOX (gBox), menuBar, FALSE, TRUE, 0);
 gtk_widget_show (menuBar);
 /* Map */
 menu = gtk_menu_new ();
 CreateMenuItem (menu, "_Open...", accelGroup, GDK_O, GDK_CONTROL_MASK, G_CALLBACK (OnMenuLoad), NULL);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "Map _limits...", accelGroup, GDK_L, GDK_CONTROL_MASK, G_CALLBACK (OnMenuLimits), NULL);
 CreateMenuItem (menu, "_Centre on...", accelGroup, GDK_C, GDK_CONTROL_MASK, G_CALLBACK (OnMenuCentre), NULL);
 AddMenuToBar (menu, "_Map", menuBar);
 /* Drawing */
 menu = gtk_menu_new ();
 group = NULL;
 mWinData . menuDrawRecords = CreateMenuItemRadio (menu, "View _records", accelGroup, GDK_R, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuViewMode), NULL, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (mWinData . menuDrawRecords));
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords), TRUE);
 mWinData . menuDrawCells = CreateMenuItemRadio (menu, "View ce_lls", accelGroup, GDK_L, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuViewMode), NULL, group);
 gtk_widget_set_sensitive (mWinData . menuDrawCells, FALSE);
 CreateMenuSeparator (menu);
 CreateMenuItem (menu, "_Record radius...", accelGroup, GDK_R, GDK_CONTROL_MASK, G_CALLBACK (OnMenuRadius), NULL);
 CreateMenuSeparator (menu);
 mWinData . menuDrawAll = CreateMenuItemCheck (menu, "_All records", accelGroup, GDK_A, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 mWinData . menuDrawRemoved = CreateMenuItemCheck (menu, "Re_moved records", accelGroup, GDK_M, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 mWinData . menuDrawBarrier = CreateMenuItemCheck (menu, "Barrier", accelGroup, GDK_B, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 CreateMenuSeparator (menu);
 mWinData . menuDrawGrid = CreateMenuItemCheck (menu, "_Grid", accelGroup, GDK_G, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuDrawCheck), NULL);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawGrid), FALSE);
 gtk_widget_set_sensitive (mWinData . menuDrawBarrier, FALSE);
 gtk_widget_set_sensitive (mWinData . menuDrawRemoved, FALSE);
 AddMenuToBar (menu, "_Drawing", menuBar);
 /* Barrier */
 menu = gtk_menu_new ();
 group = NULL;
 mWinData . menuBarrierVoronoi = CreateMenuItemRadio (menu, "_Voronoi lines", accelGroup, GDK_V, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuBarrierMode), MENU_VORONOI, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (mWinData . menuBarrierVoronoi));
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierVoronoi), TRUE);
 mWinData . menuBarrierDelaunay = CreateMenuItemRadio (menu, "_Delaunay triangles", accelGroup, GDK_D, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuBarrierMode), MENU_DELAUNAY, group);
 group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (mWinData . menuBarrierDelaunay));
 mWinData . menuBarrierMidTri = CreateMenuItemRadio (menu, "Mid tr_iangles", accelGroup, GDK_I, GDK_CONTROL_MASK,
                                             G_CALLBACK (OnMenuBarrierMode), MENU_MIDTRI, group);
 CreateMenuSeparator (menu);
 mWinData . menuBarrierAncestor = CreateMenuItemCheck (menu, "Barriers on a_ncestors", accelGroup, GDK_N, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuBarrierOptions), NULL);
 mWinData . menuBarrierCommon = CreateMenuItemCheck (menu, "Common barr_iers (on consensus)", accelGroup, GDK_I, GDK_CONTROL_MASK,
                                                    G_CALLBACK (OnMenuBarrierOptions), NULL);
 mWinData . menuBarrier = AddMenuToBar (menu, "_Barriers", menuBar);
 gtk_widget_set_sensitive (mWinData . menuBarrier, FALSE);

 /* --- Drawing area --- */
 mWinData . drawable = gtk_drawing_area_new ();
 gtk_box_pack_start (GTK_BOX (gBox), mWinData . drawable, TRUE, TRUE, 0);
 gtk_widget_show (mWinData . drawable);
 g_signal_connect (G_OBJECT (mWinData . drawable), "expose_event", G_CALLBACK (OnExpose), NULL);
 g_signal_connect (G_OBJECT (mWinData . drawable), "configure_event", G_CALLBACK (OnConfigure), NULL);
 gtk_widget_add_events (mWinData . drawable, GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK);
 g_signal_connect (G_OBJECT (mWinData . drawable), "motion_notify_event", G_CALLBACK (OnMotion), NULL);
 g_signal_connect (G_OBJECT (mWinData . drawable), "button_press_event", G_CALLBACK (OnMouseButton), NULL);

 mWinData . labelMousePos = gtk_label_new ("mouse off map");
 gtk_label_set_justify (GTK_LABEL (mWinData . labelMousePos), GTK_JUSTIFY_RIGHT);
 gtk_box_pack_start (GTK_BOX (gBox), mWinData . labelMousePos, FALSE, TRUE, 0);
 gtk_widget_show (mWinData . labelMousePos);

 gtk_widget_show (mapWindow);

 mWinData . blackPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . blackPen, mWinData . drawable -> style -> black_gc);

 mWinData . greenPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . greenPen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . greenPen, &greenColor);

 mWinData . bluePen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . bluePen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . bluePen, &blueColor);

 mWinData . redPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . redPen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . redPen, &redColor);

 mWinData . greyPen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . greyPen, mWinData . drawable -> style -> black_gc);
 gdk_gc_set_rgb_fg_color (mWinData . greyPen, &greyColor);

 mWinData . whitePen = gdk_gc_new (mWinData . drawable -> window);
 gdk_gc_copy (mWinData . whitePen, mWinData . drawable -> style -> white_gc);

 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawAll), TRUE);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved), TRUE);
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords), TRUE);

 mWinData . actNode = NULL;
 mWinData . barrierMode = DRAW_AS_VORONOI;
 mWinData . viewMode = VIEW_RECORDS;
}

void ChangeGeo () {
 PrepareGeoRefs (FALSE);
 if (GETGEOREF == NULL) {
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawGrid), FALSE);
 }
 if (activeNode != mWinData . actNode)
   mWinData . actNode = activeNode;
 DrawMap ();
}

void ActivateGrid (SFLAG mode) {
 gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawGrid), mode);
 gtk_widget_set_sensitive (mWinData . menuDrawBarrier, mode);
 gtk_widget_set_sensitive (mWinData . menuDrawRemoved, mode);
 gtk_widget_set_sensitive (mWinData . menuDrawCells, mode);
 if (mode == FALSE) {
   mWinData . viewMode = VIEW_RECORDS;
   gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords), TRUE);
 }
 DrawMap ();
}

void SetEdition (SFLAG mode) {
 if (mode == TRUE)
   mWinData . mode = EDIT_MAP_MODE;
 else
   mWinData . mode = DRAW_MAP_MODE;
}

void PaintNewGeoRef (GEOREF* geo) {
 geo -> actX = mWinData . xScale * geo -> azimuth;
 geo -> actY = mWinData . yScale * geo -> inclination;
 DrawMap ();
}

gint OnDelete (GtkWidget* treeWin, GdkEvent* event, gpointer data) {
 gtk_window_iconify (GTK_WINDOW (mapWindow));
 return TRUE;
}

gint OnExpose (GtkWidget* drawable, GdkEventExpose* event, gpointer data) {
 gdk_draw_pixmap (mWinData . drawable -> window, mWinData . drawable -> style -> fg_gc [GTK_WIDGET_STATE (mWinData . drawable)],
     mWinData . pixMap, event -> area . x, event -> area . y, event -> area . x, event -> area . y,
     event -> area . width, event -> area . height);
 return FALSE;
}

gint OnConfigure (GtkWidget* drawable, GdkEvent* event, gpointer data) {
 mWinData . clientX = mWinData . drawable -> allocation . width;
 mWinData . clientY = mWinData . drawable -> allocation . height;
 if (mWinData . pixMap)
   gdk_pixmap_unref (mWinData . pixMap);
 mWinData . pixMap = gdk_pixmap_new (mWinData . drawable -> window, mWinData . clientX, mWinData . clientY, -1);
 DrawMap ();
 return TRUE;
}

gint OnMotion (GtkWidget* drawable, GdkEventMotion* event, gpointer data) {
 SINTEGER nu;
 SREAL nuLon, nuLat;
 gchar* text;

 if (event -> state & GDK_BUTTON1_MASK) {
   if ((mWinData . prevX != 0) || (mWinData . prevY != 0)) {
     nu = event -> x - mWinData . prevX;
     mWinData . originX += nu;
     if (mWinData . originX < - (mWinData . width - 10))
       mWinData . originX = - (mWinData . width - 10);
     else if (mWinData . originX > (mWinData . clientX - 10))
       mWinData . originX = mWinData . clientX - 10;
     nu = event -> y - mWinData . prevY;
     mWinData . originY += nu;
     if (mWinData . originY < - (mWinData . height - 10))
       mWinData . originY = - (mWinData . height - 10);
     else if (mWinData . originY > (mWinData . clientY - 10))
       mWinData . originY = mWinData . clientY - 10;
     DrawMap ();
   }
   mWinData . prevX = event -> x;
   mWinData . prevY = event -> y;
 }
 else {
   mWinData . prevX = 0;
   mWinData . prevY = 0;
   nuLon = (SREAL) (event -> x - mWinData . originX) / mWinData . xScale;
   nuLat = (SREAL) (event -> y - mWinData . originY) / mWinData . yScale;
   if ((nuLon < 0) || (nuLon > 360) || (nuLat < 0) || (nuLat > 180))
     gtk_label_set_text (GTK_LABEL (mWinData . labelMousePos), "mouse off map");
   else {
     nuLat = FROMINCTOLAT (nuLat);
     nuLon = FROMAZMTOLON (nuLon);
     text = g_strdup_printf ("lat: %.4f\tlon: %.4f", nuLat, nuLon);
     gtk_label_set_text (GTK_LABEL (mWinData . labelMousePos), text);
     g_free (text);
   }
 }
 return TRUE;
}

gint OnMouseButton (GtkWidget* drawable, GdkEventButton* event, gpointer data) {
 GEOREF* actGeo;
 TERMINAL* term;
 GEOREFBROWSERDLG* browser;
 SREAL nuLon, nuLat;

 if (activeNode == NULL) return FALSE;
 if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 1)) {
   actGeo = SearchNearestPoint (event -> x - mWinData . originX, event -> y - mWinData . originY);
   if (actGeo != NULL)
     browser = GeoRefBrowser (actGeo, FALSE);
 }
 else if ((event -> type == GDK_BUTTON_PRESS) && (event -> button == 3) && (event -> state & GDK_CONTROL_MASK)) {
   actGeo = SearchNearestPoint (event -> x - mWinData . originX, event -> y - mWinData . originY);
   if (actGeo != NULL) {
     term = (TERMINAL*) actGeo -> sp -> data;
     if ((mWinData . mode == EDIT_MAP_MODE) && (term -> node == activeNode))
       browser = GeoRefBrowser (actGeo, TRUE);
     else
       browser = GeoRefBrowser (actGeo, FALSE);
   }
 }
 else if ((mWinData . mode == EDIT_MAP_MODE) && (event -> button == 3) && (event -> state & GDK_SHIFT_MASK)
          && (g_node_first_child (activeNode) == NULL)) {
   nuLon = (SREAL) (event -> x - mWinData . originX) / mWinData . xScale;
   nuLat = (SREAL) (event -> y - mWinData . originY) / mWinData . yScale;
   nuLat = FROMINCTOLAT (nuLat);
   nuLon = FROMAZMTOLON (nuLon);
   if ((nuLon > -180.00001) && (nuLon < 180.00001) && (nuLat > -90.00001) && (nuLat < 90.00001)) {
     actGeo = AddGeoRef (nuLon, nuLat, activeNode);
     PaintNewGeoRef (actGeo);
   }
 }
 return FALSE;
}

void OnMenuLoad (GtkMenuItem* openMenu, gpointer data) {
 GtkWidget* dialog;
 GtkFileFilter* filter;
 GdkPixbuf* nuMap;
 gchar* fileName;

 fileName = NULL;
 dialog = gtk_file_chooser_dialog_new ("Open Map", GTK_WINDOW (GETMAINWIN), GTK_FILE_CHOOSER_ACTION_OPEN,
                                       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
 filter = gtk_file_filter_new ();
 gtk_file_filter_add_pixbuf_formats (filter);
 gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog), filter);
 if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
   fileName = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
 gtk_widget_destroy (dialog);
 if (fileName != NULL) {
   nuMap = gdk_pixbuf_new_from_file (fileName, NULL);
   if (nuMap != NULL) {
     mWinData . mapImg = nuMap;
     mWinData . width = gdk_pixbuf_get_width (nuMap);
     mWinData . height = gdk_pixbuf_get_height (nuMap);
     mWinData . originX = mWinData . originY = 0;
     mWinData . startX = mWinData . startY = 0;
     mWinData . north = mWinData . west = 0;
     mWinData . south = 180.0000;
     mWinData . east = 360.0000;
     PrepareGeoRefs (TRUE);
     DrawMap ();
   }
 }
 g_free (fileName);
}

void OnMenuLimits (GtkMenuItem* limitsMenu, gpointer data) {
 MAPLIMITSDIALOG* mapLimitsDlg;
 gchar* entry;
 SREAL tmp, nuN, nuS, nuW, nuE;

 mapLimitsDlg = MapLimitsDialog (FROMINCTOLAT (mWinData . north), FROMINCTOLAT (mWinData . south), FROMAZMTOLON (mWinData . east),
                                 FROMAZMTOLON (mWinData . west));
 nuN = mWinData . north;
 nuS = mWinData . south;
 nuE = mWinData . east;
 nuW = mWinData . west;
 if (gtk_dialog_run (GTK_DIALOG (mapLimitsDlg -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> northEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -90.0000) && (tmp < 90.0000))
       nuN = FROMLATTOINC (tmp);
     else
       nuN = mWinData . north;
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> southEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -90.0000) && (tmp < 90.0000))
       nuS = FROMLATTOINC (tmp);
     else
       nuS = mWinData . south;
   }
   if (nuS < nuN) {
     nuN = mWinData . north;
     nuS = mWinData . south;
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> westEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -180.0000) && (tmp < 180.0000))
       nuW = FROMLONTOAZM (tmp);
     else
       nuW = mWinData . west;
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (mapLimitsDlg -> eastEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&tmp) != FALSE) {
     if ((tmp > -180.0000) && (tmp < 180.0000))
       nuE = FROMLONTOAZM (tmp);
     else
       nuE = mWinData . east;
   }
   if (nuE < nuW) {
     nuW = mWinData . west;
     nuE = mWinData . east;
   }
 }
 DestroyMapLimitsDialog (mapLimitsDlg);
 if ((nuN != mWinData . north) || (nuS != mWinData . south) || (nuE != mWinData . east) || (nuW != mWinData . west)) {
  mWinData . north = nuN;
  mWinData . south = nuS;
  mWinData . east = nuE;
  mWinData . west = nuW;
  PrepareGeoRefs (TRUE);
  mWinData . startX = mWinData . xScale * mWinData . west;
  mWinData . startY = mWinData . yScale * mWinData . north;
  DrawMap ();
 }
}

void OnMenuCentre (GtkMenuItem* centreMenu, gpointer data) {
 CENTREDIALOG* centreDialog;
 gchar* entry;
 SREAL nuX, nuY, cLat, cLon;
 SFLAG change;

 centreDialog = CentreDialog ();
 change = FALSE;
 if (gtk_dialog_run (GTK_DIALOG (centreDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (centreDialog -> latEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdReal (&cLat) != FALSE) {
     if ((cLat > -90.0000) && (cLat < 90.0000)) {
       cLat = FROMLATTOINC (cLat);
       change = TRUE;
     }
   }
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (centreDialog -> lonEntry)));
   SetCommandLine (entry);
   g_free (entry);
   if ((GetCmdReal (&cLon) != FALSE) && (change == TRUE)) {
     if ((cLon > -180.0000) && (cLon < 180.0000))
       cLon = FROMLONTOAZM (cLon);
     else
       change = FALSE;
   }
   else
     change = FALSE;
 }
 DestroyCentreDialog (centreDialog);
 if (change == TRUE) {
   nuX = mWinData . xScale * cLon;
   nuY = mWinData . yScale * cLat;
   mWinData . originX = (mWinData . clientX / 2) - nuX;
   mWinData . originY = (mWinData . clientY / 2) - nuY;
   DrawMap ();
 }
}

void OnMenuRadius (GtkMenuItem* radiusMenu, gpointer data) {
 COUNTERENTRYDIALOG* radiusDialog;
 gchar* entry;
 SCOUNTER nuRadius;
 SFLAG paint;

 radiusDialog = CounterEntryDialog (GTK_WINDOW (mapWindow), mWinData . radius, "Set drawing radius", "Radius");
 paint = FALSE;
 if (gtk_dialog_run (GTK_DIALOG (radiusDialog -> dialog)) == GTK_RESPONSE_ACCEPT) {
   entry = g_strdup_printf (gtk_entry_get_text (GTK_ENTRY (radiusDialog -> entry)));
   SetCommandLine (entry);
   g_free (entry);
   if (GetCmdNumber (&nuRadius) != FALSE) {
     if ((nuRadius >= 0) && (nuRadius != mWinData . radius)) {
       mWinData . radius = nuRadius;
       paint = TRUE;
     }
   }
 }
 DestroyCounterEntryDialog (radiusDialog);
 if (paint == TRUE)
   DrawMap ();
}

void OnMenuViewMode (GtkMenuItem* modeMenu, gpointer data) {
 SFLAG prevMode;

 prevMode = mWinData . viewMode;
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRecords)) == TRUE)
   mWinData . viewMode = VIEW_RECORDS;
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawCells)) == TRUE)
   mWinData . viewMode = VIEW_CELLS;
 if (prevMode != mWinData . viewMode)
   DrawMap ();
}

void OnMenuDrawCheck (GtkMenuItem* drawCheckMenu, gpointer data) {
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawBarrier)) == TRUE)
   gtk_widget_set_sensitive (mWinData . menuBarrier, TRUE);
 else
   gtk_widget_set_sensitive (mWinData . menuBarrier, FALSE);
 DrawMap ();
}

void OnMenuBarrierMode (GtkMenuItem* modeMenu, gpointer data) {
 SFLAG prevMode;

 prevMode = mWinData . barrierMode;
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierVoronoi)) == TRUE)
   mWinData . barrierMode = DRAW_AS_VORONOI;
 else if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierDelaunay)) == TRUE)
   mWinData . barrierMode = DRAW_AS_DELAUNAY;
 else if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierMidTri)) == TRUE)
   mWinData . barrierMode = DRAW_AS_MIDTRI;
 if (prevMode != mWinData . barrierMode)
   DrawMap ();
}

void OnMenuBarrierOptions (GtkMenuItem* barrierMenu, gpointer data) {
 DrawMap ();
}

void PrepareGeoRefs (SFLAG updateAll) {
 GList* navGeo;
 GEOREF* ref;

 mWinData . xScale = (SREAL) mWinData . width / (mWinData . east - mWinData . west);
 mWinData . yScale = (SREAL) mWinData . height / (mWinData . south - mWinData . north);
 for (navGeo = GETLASTGEOREF; navGeo != NULL; navGeo = g_list_previous (navGeo)) {
   ref = (GEOREF*) navGeo -> data;
   if ((updateAll == FALSE) && (ref -> actX != -1)) break;
   ref -> actX = mWinData . xScale * ref -> azimuth;
   ref -> actY = mWinData . yScale * ref -> inclination;
 }
}

void DrawMap () {
 GdkRectangle updateRec;
 GList* navGeo;
 GEOREF* ref;
 PHYLONODE* node;
 GNode* navAnc;
 PHYLONODE* ancNode;
 SINTEGER x, y;
 SCOUNTER r, c, sX, sY, eX, eY;
 GdkGCValues gcVals;
 SREAL tmpNum;

 if (mWinData . pixMap == NULL)
   mWinData . pixMap = gdk_pixmap_new (mWinData . drawable -> window, mWinData . clientX, mWinData . clientY, -1);
 gdk_draw_rectangle (mWinData . pixMap, mWinData . drawable -> style -> white_gc, TRUE, 0, 0, mWinData . clientX, mWinData . clientY);
 if (mWinData . mapImg != NULL) {
   gdk_draw_pixbuf (mWinData . pixMap, mWinData . drawable -> style -> fg_gc [GTK_WIDGET_STATE (mWinData . drawable)], mWinData . mapImg,
                    0, 0, mWinData . originX + mWinData . startX, mWinData . originY + mWinData . startY,
                    -1, -1, GDK_RGB_DITHER_NONE, 0, 0);
 }
 /* Draw the grid */
 if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawGrid)) == TRUE) {
   sX = mWinData . originX;
   eX = (360 * mWinData . xScale) + mWinData . originX;
   if (((sX >= 0) && (sX <= mWinData . clientX)) || ((eX >= 0) && (eX <= mWinData . clientX)) ||
       ((sX <= 0) && (eX >= mWinData . clientX))) {
     if (sX < 0)
       sX = 0;
     if (eX > mWinData . clientX)
       eX = mWinData . clientX;
     for (r = 0; ; ++ r) {
       tmpNum = GETGRIDSIZE * r;
       if (tmpNum > 180) break;
       sY = (tmpNum * mWinData . yScale) + mWinData . originY;
       if (sY < 0) continue;
       if (sY > mWinData . clientY) break;
       gdk_draw_line (mWinData . pixMap, mWinData . greyPen, sX, sY, eX, sY);
     }
   }
   sY = mWinData . originY;
   eY = (180 * mWinData . yScale) + mWinData . originY;
   if (((sY >= 0) && (sY <= mWinData . clientY)) || ((eY >= 0) && (eY <= mWinData . clientY)) ||
       ((sY <= 0) && (eY >= mWinData . clientY))) {
     if (sY < 0)
       sY = 0;
     if (eY > mWinData . clientY)
       eY = mWinData . clientY;
     for (c = 0; ; ++ c) {
       tmpNum = GETGRIDSIZE * c;
       if (tmpNum > 360) break;
       sX = (tmpNum * mWinData . xScale) + mWinData . originX;
       if (sX < 0) continue;
       if (sX > mWinData . clientX) break;
       gdk_draw_line (mWinData . pixMap, mWinData . greyPen, sX, sY, sX, eY);
     }
   }
 }
 /* First draw all records */
 if ((gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawAll)) == TRUE) &&
     (mWinData . viewMode == VIEW_RECORDS)) {
   for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
     ref = (GEOREF*) navGeo -> data;
     x = ref -> actX + mWinData . originX;
     if ((x < 0) || (x > mWinData . clientX)) continue;
     y = ref -> actY + mWinData . originY;
     if ((y < 0) || (y > mWinData . clientY)) continue;
     gdk_draw_rectangle (mWinData . pixMap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
   }
 }
 /* Then draws the active descendants of the actual node */
 if (activeNode != NULL) {
   node = (PHYLONODE*) activeNode -> data;
   if (node -> rec == NULL)
     DrawDescs (activeNode, mWinData . greenPen);
   else if (node -> rec -> nodeFlag & VIC_NODE) {
     DrawVicRec (node -> rec -> setA -> node, mWinData . bluePen, node -> rec -> setB -> rec -> inher);
     DrawVicRec (node -> rec -> setB -> node, mWinData . redPen, node -> rec -> setA -> rec -> inher);
     if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawBarrier)) == TRUE) {
       gdk_gc_get_values (mWinData . greenPen, &gcVals);
       gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width, GDK_LINE_ON_OFF_DASH, gcVals . cap_style,
                                   gcVals . join_style);
       if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierAncestor)) == TRUE) {
         for (navAnc = node -> node -> parent; navAnc != NULL; navAnc = navAnc -> parent) {
           ancNode = (PHYLONODE*) navAnc -> data;
           if (ancNode -> rec -> nodeFlag & VIC_NODE)
             DrawBarrierFromRec (ancNode -> rec);
         }
       }
       gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width, gcVals . line_style, gcVals . cap_style,
                                   gcVals . join_style);
       if ((gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuBarrierCommon)) == TRUE) &&
           (node -> alloSet != NULL))
         DrawBarrierFromAlloSet (node);
       gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width + 2, gcVals . line_style, gcVals . cap_style,
                                   gcVals . join_style);
       DrawBarrierFromRec (node -> rec);
       gdk_gc_set_line_attributes (mWinData . greenPen, gcVals . line_width, gcVals . line_style, gcVals . cap_style,
                                   gcVals . join_style);
     }
   }
   else
     DrawDescsRec (activeNode, mWinData . greenPen);
 }
 updateRec . x = 0;
 updateRec . y = 0;
 updateRec . width = mWinData . clientX;
 updateRec . height = mWinData . clientY;
 gdk_window_invalidate_rect (mWinData . drawable -> window, &updateRec, TRUE);
}

void DrawDescs (GNode* refNode, GdkGC* pen) {
 GNode* actNode;
 GList* navGeo;
 PHYLONODE* node;
 TERMINAL* term;
 GEOREF* ref;
 SINTEGER x, y;

 actNode = refNode;
 if (g_node_first_child (actNode) != NULL)
   actNode = GetNextTerm (actNode);
 for (; actNode != NULL; actNode = GetNextTerm (actNode)) {
   if (IsDesc (actNode, refNode) == TRUE) {
     node = (PHYLONODE*) actNode -> data;
     if (node -> rec != NULL) {}
     term = node -> term;
     for (navGeo = term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       ref = (GEOREF*) navGeo -> data;
       if (term != (TERMINAL*) ref -> sp -> data) break;
       x = ref -> actX + mWinData . originX;
       if ((x < 0) || (x > mWinData . clientX)) continue;
       y = ref -> actY + mWinData . originY;
       if ((y < 0) || (y > mWinData . clientY)) continue;
       gdk_draw_rectangle (mWinData . pixMap, pen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
       gdk_draw_rectangle (mWinData . pixMap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
     }
   }
   else break;
 }
}

void DrawVicRec (GNode* refNode, GdkGC* pen, BITFIELD* outCells) {
 GNode* actNode;
 GNode* ancNode;
 GList* navGeo;
 PHYLONODE* node;
 PHYLONODE* anc;
 TERMINAL* term;
 GEOREF* ref;
 SCOUNTER c, sX, sY;
 SINTEGER x, y, x1, y1;
 BITFIELD* remov;
 BITFIELD* inCells;
 SFLAG in, out, rem;

 node = (PHYLONODE*) refNode -> data;
 if (mWinData . viewMode == VIEW_RECORDS) {
   remov = CreateBitField ();
   inCells = node -> rec -> inher;
   for (actNode = refNode; actNode != NULL;) {
     if (IsDesc (actNode, refNode) != TRUE) break;
     node = (PHYLONODE*) actNode -> data;
     if (node -> rec -> numInher == 0) {
     /* A full j-node, ignores all of its descendants */
       while (actNode != NULL) {
         if (g_node_next_sibling (actNode) != NULL) {
           actNode = g_node_next_sibling (actNode);
           break;
         }
         actNode = actNode -> parent;
       }
       continue;
     }
     if (g_node_first_child (actNode) != NULL) {
     /* Not a term, goes to the first descendant */
       actNode = g_node_first_child (actNode);
       continue;
     }
     /* Updates the removed cells */
     ResetBitField (remov);
     for (ancNode = actNode; ; ancNode = ancNode -> parent) {
       anc = (PHYLONODE*) ancNode -> data;
       BitFieldORsum (remov, anc -> rec -> remov);
       if (ancNode == refNode) break;
     }
     term = node -> term;
     for (navGeo = term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       ref = (GEOREF*) navGeo -> data;
       if (term != (TERMINAL*) ref -> sp -> data) break;
       x = ref -> actX + mWinData . originX;
       if ((x < 0) || (x > mWinData . clientX)) continue;
       y = ref -> actY + mWinData . originY;
       if ((y < 0) || (y > mWinData . clientY)) continue;
       rem = IsBitOn (remov, ref -> actCell);
       in = IsBitOn (inCells, ref -> actCell);
       out = IsBitOn (outCells, ref -> actCell);
       if ((rem == TRUE) || (in != TRUE)) {
         if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved)) == TRUE) {
           gdk_draw_rectangle (mWinData . pixMap, mWinData . whitePen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
           gdk_draw_rectangle (mWinData . pixMap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         }
       }
       else if ((in == TRUE) && (out != TRUE)) {
         gdk_draw_rectangle (mWinData . pixMap, pen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         gdk_draw_rectangle (mWinData . pixMap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
       }
       else {
         gdk_draw_rectangle (mWinData . pixMap, mWinData . greenPen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         gdk_draw_rectangle (mWinData . pixMap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
       }
     }
     while (actNode != NULL) {
       if (g_node_next_sibling (actNode) != NULL) {
         actNode = g_node_next_sibling (actNode);
         break;
       }
       actNode = actNode -> parent;
     }
   }
   g_slice_free1 (GETBITFIELDSIZE, remov);
 }
 else {
   for (c = 0; c < GETNUMCELLS; ++ c) {
     if (IsBitOn (node -> rec -> cells, c)) {
       GetColRowFromCell (c, &sY, &sX);
       x = (((GETGRIDSIZE * sX) + GETGRIDSTARTX) * mWinData . xScale) + mWinData . originX;
       x1 = x + (GETGRIDSIZE * mWinData . xScale);
       if ((x1 < 0) || (x > mWinData . clientX)) continue;
       y = (((GETGRIDSIZE * sY) + GETGRIDSTARTY) * mWinData . yScale) + mWinData . originY;
       y1 = y + (GETGRIDSIZE * mWinData . yScale);
       if ((y1 < 0) || (y > mWinData . clientY)) continue;
       if (IsBitOn (outCells, c))
         gdk_draw_rectangle (mWinData . pixMap, mWinData . greenPen, TRUE, x, y, (GETGRIDSIZE * mWinData . xScale),
                             (GETGRIDSIZE * mWinData . yScale));
       else
         gdk_draw_rectangle (mWinData . pixMap, pen, TRUE, x, y, (GETGRIDSIZE * mWinData . xScale),
                             (GETGRIDSIZE * mWinData . yScale));
     }
   }
 }
}

void DrawDescsRec (GNode* refNode, GdkGC* pen) {
 GNode* actNode;
 GNode* ancNode;
 GList* navGeo;
 PHYLONODE* node;
 PHYLONODE* anc;
 TERMINAL* term;
 GEOREF* ref;
 SCOUNTER c, sX, sY;
 SINTEGER x, y, x1, y1;
 BITFIELD* remov;
 BITFIELD* inCells;
 SFLAG in, rem;

 node = (PHYLONODE*) refNode -> data;
 if (mWinData . viewMode == VIEW_RECORDS) {
   remov = CreateBitField ();
   inCells = node -> rec -> cells;
   for (actNode = refNode; actNode != NULL;) {
     if (IsDesc (actNode, refNode) != TRUE) break;
     node = (PHYLONODE*) actNode -> data;
     if ((actNode != refNode) && (node -> rec -> numInher == 0)) {
     /* A full j-node, ignores all of its descendants */
       while (actNode != NULL) {
         if (g_node_next_sibling (actNode) != NULL) {
           actNode = g_node_next_sibling (actNode);
           break;
         }
         actNode = actNode -> parent;
       }
       continue;
     }
     if (g_node_first_child (actNode) != NULL) {
     /* Not a term, goes to the first descendant */
       actNode = g_node_first_child (actNode);
       continue;
     }
     /* Updates the removed cells */
     ResetBitField (remov);
     for (ancNode = actNode; ; ancNode = ancNode -> parent) {
       anc = (PHYLONODE*) ancNode -> data;
       if (ancNode == refNode) break;
       BitFieldORsum (remov, anc -> rec -> remov);
     }
     term = node -> term;
     for (navGeo = term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       ref = (GEOREF*) navGeo -> data;
       if (term != (TERMINAL*) ref -> sp -> data) break;
       x = ref -> actX + mWinData . originX;
       if ((x < 0) || (x > mWinData . clientX)) continue;
       y = ref -> actY + mWinData . originY;
       if ((y < 0) || (y > mWinData . clientY)) continue;
       rem = IsBitOn (remov, ref -> actCell);
       in = IsBitOn (inCells, ref -> actCell);
       if ((rem == TRUE) || (in != TRUE)) {
         if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mWinData . menuDrawRemoved)) == TRUE) {
           gdk_draw_rectangle (mWinData . pixMap, mWinData . whitePen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
           gdk_draw_rectangle (mWinData . pixMap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         }
       }
       else if (in == TRUE) {
         gdk_draw_rectangle (mWinData . pixMap, pen, TRUE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
         gdk_draw_rectangle (mWinData . pixMap, mWinData . blackPen, FALSE, x - mWinData . radius, y - mWinData . radius,
                 (2 * mWinData . radius) + 1, (2 * mWinData . radius) + 1);
       }
     }
     while (actNode != NULL) {
       if (g_node_next_sibling (actNode) != NULL) {
         actNode = g_node_next_sibling (actNode);
         break;
       }
       actNode = actNode -> parent;
     }
   }
   g_slice_free1 (GETBITFIELDSIZE, remov);
 }
 else {
   for (c = 0; c < GETNUMCELLS; ++ c) {
     if (IsBitOn (node -> rec -> cells, c)) {
       GetColRowFromCell (c, &sY, &sX);
       x = (((GETGRIDSIZE * sX) + GETGRIDSTARTX) * mWinData . xScale) + mWinData . originX;
       x1 = x + (GETGRIDSIZE * mWinData . xScale);
       if ((x1 < 0) || (x > mWinData . clientX)) continue;
       y = (((GETGRIDSIZE * sY) + GETGRIDSTARTY) * mWinData . yScale) + mWinData . originY;
       y1 = y + (GETGRIDSIZE * mWinData . yScale);
       if ((y1 < 0) || (y > mWinData . clientY)) continue;
       gdk_draw_rectangle (mWinData . pixMap, pen, TRUE, x, y, (GETGRIDSIZE * mWinData . xScale), (GETGRIDSIZE * mWinData . yScale));
     }
   }
 }
}

GEOREF* SearchNearestPoint (gint x, gint y) {
 GList* navGeo;
 GEOREF* actGeo;
 GEOREF* toRet;
 SINTEGER difX, difY, bestDif, actDif;
 PHYLONODE* node;

 toRet = NULL;
 bestDif = (2 * mWinData . radius) + 1;
 /* First look for points of the active terminal, if any */
 if (activeNode != NULL) {
   node = (PHYLONODE*) activeNode -> data;
   if (node -> term != NULL) {
     for (navGeo = node -> term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       actGeo = (GEOREF*) navGeo -> data;
       if (actGeo -> sp != node -> term -> list) break;
       difX = x - actGeo -> actX;
       difX = (difX >= 0) ? difX : - difX;
       if (difX > (mWinData . radius + 1)) continue;
       difY = y - actGeo -> actY;
       difY = (difY >= 0) ? difY : - difY;
       if (difY > (mWinData . radius + 1)) continue;
       actDif = difX + difY;
       if (actDif < bestDif) {
         bestDif = actDif;
         toRet = actGeo;
       }
       if (actDif == 0) break;
     }
   }
 }
 if (toRet != NULL) return toRet;
 /* Then, check all points */
 for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
   actGeo = (GEOREF*) navGeo -> data;
   difX = x - actGeo -> actX;
   difX = (difX >= 0) ? difX : - difX;
   if (difX > (mWinData . radius + 1)) continue;
   difY = y - actGeo -> actY;
   difY = (difY >= 0) ? difY : - difY;
   if (difY > (mWinData . radius + 1)) continue;
   actDif = difX + difY;
   if (actDif < bestDif) {
     bestDif = actDif;
     toRet = actGeo;
   }
   if (actDif == 0) break;
 }
 return toRet;
}

void DrawBarrierFromRec (RECONS* rec) {
 BITFIELD* setA;
 BITFIELD* setB;

 if (rec -> triangles == NULL) {
   setA = CreateBitField ();
   setB = CreateBitField ();
   GetObsSet (setA, rec -> setA -> node);
   GetObsSet (setB, rec -> setB -> node);
   BitFieldANDsum (setA, rec -> setA -> rec -> inher);
   BitFieldANDsum (setB, rec -> setB -> rec -> inher);
   BitFieldRemove (setA, rec -> setB -> rec -> inher);
   BitFieldRemove (setB, rec -> setA -> rec -> inher);
   rec -> triangles = MakeVoronoi (setA, setB);
   g_slice_free1 (GETBITFIELDSIZE, setA);
   g_slice_free1 (GETBITFIELDSIZE, setB);
 }
 DrawBarrier (rec -> triangles);
}

void DrawBarrierFromAlloSet (PHYLONODE* node) {
 if (node -> alloSet -> triangles == NULL)
   node -> alloSet -> triangles = MakeVoronoi (node -> alloSet -> setA, node -> alloSet -> setB);
 DrawBarrier (node -> alloSet -> triangles);
}

void DrawBarrier (GPtrArray* triangles) {
 SCOUNTER i;
 SCOUNTER aR, aC, bR, bC, cR, cC;
 SCOUNTER sX, sY, eX, eY, mid;
 SCOUNTER cX, cY, nX, nY;
 DELAUNAYTRIANGLE* actTri;
 DELAUNAYTRIANGLE* neiTri;

 for (i = 0; i < triangles -> len; ++ i) {
   actTri = (DELAUNAYTRIANGLE*) g_ptr_array_index (triangles, i);
   GetColRowFromCell (actTri -> cellA, &aR, &aC);
   GetColRowFromCell (actTri -> cellB, &bR, &bC);
   GetColRowFromCell (actTri -> cellC, &cR, &cC);
   aC = (GETGRIDSTARTX + (GETGRIDSIZE * aC) + (GETGRIDSIZE / 2)) * mWinData . xScale;
   bC = (GETGRIDSTARTX + (GETGRIDSIZE * bC) + (GETGRIDSIZE / 2)) * mWinData . xScale;
   cC = (GETGRIDSTARTX + (GETGRIDSIZE * cC) + (GETGRIDSIZE / 2)) * mWinData . xScale;
   aR = (GETGRIDSTARTY + (GETGRIDSIZE * aR) + (GETGRIDSIZE / 2)) * mWinData . yScale;
   bR = (GETGRIDSTARTY + (GETGRIDSIZE * bR) + (GETGRIDSIZE / 2)) * mWinData . yScale;
   cR = (GETGRIDSTARTY + (GETGRIDSIZE * cR) + (GETGRIDSIZE / 2)) * mWinData . yScale;
   sX = sY = eX = eY =-1;

   if (mWinData . barrierMode == DRAW_AS_DELAUNAY) {
     /* Draws the triangles */
     sX = aC + mWinData . originX;
     sY = aR + mWinData . originY;
     eX = bC + mWinData . originX;
     eY = bR + mWinData . originY;
     gdk_draw_line (mWinData . pixMap, mWinData . greenPen, sX, sY, eX, eY);
     eX = cC + mWinData . originX;
     eY = cR + mWinData . originY;
     gdk_draw_line (mWinData . pixMap, mWinData . greenPen, sX, sY, eX, eY);
     sX = bC + mWinData . originX;
     sY = bR + mWinData . originY;
     gdk_draw_line (mWinData . pixMap, mWinData . greenPen, sX, sY, eX, eY);
   }
   else if (mWinData . barrierMode == DRAW_AS_MIDTRI) {
     /* Draw middle triangles, "pseudovoronoi" */
     if (actTri -> stateA != actTri -> stateB) {
       mid = (aC > bC) ? ((aC - bC) / 2) + bC : ((bC - aC) / 2) + aC;
       sX = mid + mWinData . originX;
       mid = (aR > bR) ? ((aR - bR) / 2) + bR : ((bR - aR) / 2) + aR;
       sY = mid + mWinData . originY;
     }
     if (actTri -> stateA != actTri -> stateC) {
       mid = (aC > cC) ? ((aC - cC) / 2) + cC : ((cC - aC) / 2) + aC;
       eX = mid + mWinData . originX;
       mid = (aR > cR) ? ((aR - cR) / 2) + cR : ((cR - aR) / 2) + aR;
       eY = mid + mWinData . originY;
     }
     if ((actTri -> stateB != actTri -> stateC) && (sX != -1)) {
       mid = (bC > cC) ? ((bC - cC) / 2) + cC : ((cC - bC) / 2) + bC;
       eX = mid + mWinData . originX;
       mid = (bR > cR) ? ((bR - cR) / 2) + cR : ((cR - bR) / 2) + bR;
       eY = mid + mWinData . originY;
     }
     if ((actTri -> stateB != actTri -> stateC) && (sX == -1)) {
       mid = (bC > cC) ? ((bC - cC) / 2) + cC : ((cC - bC) / 2) + bC;
       sX = mid + mWinData . originX;
       mid = (bR > cR) ? ((bR - cR) / 2) + cR : ((cR - bR) / 2) + bR;
       sY = mid + mWinData . originY;
     }
     if (((sX < 0) && (eX < 0)) || ((sY < 0) && (eY < 0)) ||
        ((sX > mWinData . clientX) && (eX > mWinData . clientX)) || ((sY > mWinData . clientY) && (eY > mWinData . clientY)))
       continue;
     gdk_draw_line (mWinData . pixMap, mWinData . greenPen, sX, sY, eX, eY);
   }
   else {
     /* Draw Voronoi lines */
     cX = ((GETGRIDSTARTX + (GETGRIDSIZE * actTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + mWinData . originX;
     cY = ((GETGRIDSTARTY + (GETGRIDSIZE * actTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + mWinData . originY;
     if ((actTri -> stateA != actTri -> stateB) && (actTri -> abNeighbor > i)) {
       neiTri = (DELAUNAYTRIANGLE*) g_ptr_array_index (triangles, actTri -> abNeighbor);
       nX = ((GETGRIDSTARTX + (GETGRIDSIZE * neiTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + mWinData . originX;
       nY = ((GETGRIDSTARTY + (GETGRIDSIZE * neiTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + mWinData . originY;
       if (!(((cX < 0) && (nX < 0)) || ((cY < 0) && (nY < 0)) ||
        ((cX > mWinData . clientX) && (nX > mWinData . clientX)) || ((cY > mWinData . clientY) && (nY > mWinData . clientY))))
          gdk_draw_line (mWinData . pixMap, mWinData . greenPen, cX, cY, nX, nY);
     }
     if ((actTri -> stateA != actTri -> stateC) && (actTri -> acNeighbor > i)) {
       neiTri = (DELAUNAYTRIANGLE*) g_ptr_array_index (triangles, actTri -> acNeighbor);
       nX = ((GETGRIDSTARTX + (GETGRIDSIZE * neiTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + mWinData . originX;
       nY = ((GETGRIDSTARTY + (GETGRIDSIZE * neiTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + mWinData . originY;
       if (!(((cX < 0) && (nX < 0)) || ((cY < 0) && (nY < 0)) ||
        ((cX > mWinData . clientX) && (nX > mWinData . clientX)) || ((cY > mWinData . clientY) && (nY > mWinData . clientY))))
         gdk_draw_line (mWinData . pixMap, mWinData . greenPen, cX, cY, nX, nY);
     }
     if ((actTri -> stateB != actTri -> stateC) && (actTri -> bcNeighbor > i)) {
       neiTri = (DELAUNAYTRIANGLE*) g_ptr_array_index (triangles, actTri -> bcNeighbor);
       nX = ((GETGRIDSTARTX + (GETGRIDSIZE * neiTri -> cX) + (GETGRIDSIZE / 2)) * mWinData . xScale) + mWinData . originX;
       nY = ((GETGRIDSTARTY + (GETGRIDSIZE * neiTri -> cY) + (GETGRIDSIZE / 2)) * mWinData . yScale) + mWinData . originY;
       if (!(((cX < 0) && (nX < 0)) || ((cY < 0) && (nY < 0)) ||
        ((cX > mWinData . clientX) && (nX > mWinData . clientX)) || ((cY > mWinData . clientY) && (nY > mWinData . clientY))))
         gdk_draw_line (mWinData . pixMap, mWinData . greenPen, cX, cY, nX, nY);
     }
   }
 }
}
