#include "out.h"
#include "biogeo.h"
#include <stdio.h>

static GString* DateString (gchar* date);

void WriteXMLFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 GNode* navNode;
 GList* navGeo;
 GList* navName;
 PHYLONODE* node;
 PHYLOTREE* tree;
 GEOREF* geo;
 FILE* out;
 gchar* numStr;
 SCOUNTER c;

 iniLen = (85 * GETNUMSPP) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<Vip>");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   g_string_append (outBuffer, "\n");
   for (navNode = tree -> root; navNode != NULL; ) {
     node = (PHYLONODE*) navNode -> data;
     g_string_append_printf (outBuffer, "\n<Node id=\"%ld\">", node -> id);
     if (node -> name != NULL)
       g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
     if (g_node_first_child (navNode) != NULL) {
       navNode = g_node_first_child (navNode);
       continue;
     }
     for (c = 0; ; ++ c) {
       navName = g_list_nth (node -> term -> synonyms, c);
       if (navName == NULL) break;
       if (navName -> data == node -> term -> name) continue;
       g_string_append_printf (outBuffer, "\n\t<synonym>%s</synonym>", (gchar*) navName -> data);
     }
     for (navGeo = node -> term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       geo = (GEOREF*) navGeo -> data;
       if (geo -> sp != node -> term -> list) break;
       g_string_append (outBuffer, "\n\t<SpecimenInfo>");
       if (geo -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", geo -> name);
       if (geo -> collec != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collection>%s</collection>", geo -> collec);
       if (geo -> catalog != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<catalogCode>%s</catalogCode>", geo -> catalog);
       if (geo -> url != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<url>%s</url>", geo -> url);
       if (geo -> reference != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<reference>%s</reference>", geo -> reference);
       if (geo -> identifier != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<identifiedBy>%s</identifiedBy>", geo -> identifier);
       if (geo -> sampleDate != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<collectionDate>%s</collectionDate>", geo -> sampleDate);
       g_string_append (outBuffer, "\n\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", geo -> lon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", geo -> lat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t</Point>\n\t</SpecimenInfo>");
     }
     while (navNode != NULL) {
       g_string_append (outBuffer, "\n</Node>");
       if (g_node_next_sibling (navNode) != NULL) {
         navNode = g_node_next_sibling (navNode);
         break;
       }
       navNode = navNode -> parent;
     }
   }
 }
 g_string_append (outBuffer, "\n\n</Vip>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteKMLFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 GNode* navNode;
 GList* navGeo;
 PHYLONODE* node;
 PHYLOTREE* tree;
 GEOREF* geo;
 FILE* out;
 gchar* numStr;
 GString* date;

 iniLen = (85 * GETNUMSPP) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\"?><kml xmlns=\"http://earth.google.com/kml/2.1\">");
 g_string_append (outBuffer, "\n\n<Document>");
 g_string_append_printf (outBuffer, "\n\t<name>%s</name>", fileName);
 g_string_append (outBuffer, "\n\t<description><![CDATA[<p>KML file generated by VIP (Arias, 2009)</p>");
 g_string_append (outBuffer, "\n\t\t<p>Useful links:<br>");
 g_string_append (outBuffer, "<a href=\"http://www.zmuc.dk/public/phylogeny/Endemism/\">");
 g_string_append (outBuffer, "www.zmuc.dk/public/phylogeny/Endemism</a>: NDM download<br />");
 g_string_append (outBuffer, "\n\t\t<a href=\"mailto:sal.arias@gmail.com\">sal.arias@gmail.com</a>: ");
 g_string_append (outBuffer, "Salva e-mail</p>\n\t\t<p><b>Acknowledgment</b><br />Pablo, Claudia, ");
 g_string_append (outBuffer, "and Lolo, for suggestions and interest!</p>]]></description>\n");
 g_string_append (outBuffer, "\n");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   g_string_append (outBuffer, "\n");
   for (navNode = tree -> root; navNode != NULL; ) {
     node = (PHYLONODE*) navNode -> data;
     g_string_append_printf (outBuffer, "\n<Folder id=\"%ld\">", node -> id);
     if (node -> name != NULL)
       g_string_append_printf (outBuffer, "\n\t<name>%s</name>", node -> name);
     if (g_node_first_child (navNode) != NULL) {
       navNode = g_node_first_child (navNode);
       continue;
     }
     for (navGeo = node -> term -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
       geo = (GEOREF*) navGeo -> data;
       if (geo -> sp != node -> term -> list) break;
       g_string_append (outBuffer, "\n\t<Placemark>");
       if (geo -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t\t<name>%s</name>", geo -> name);
       g_string_append (outBuffer, "\n\t\t<description><![CDATA[<p>");
       if (geo -> name != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tTerminal: %s<br />", geo -> name);
       if (geo -> collec != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollection: %s<br />", geo -> collec);
       if (geo -> catalog != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCatalogue code: %s <br />", geo -> catalog);
       if (geo -> url != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tUrl: <a hfer=\"%s\">%s</a><br />", geo -> url, geo -> url);
       if (geo -> reference != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tReference: %s<br />", geo -> reference);
       if (geo -> identifier != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tIdentified by: %s<br />", geo -> identifier);
       if (geo -> sampleDate != NULL)
         g_string_append_printf (outBuffer, "\n\t\t\tCollection date: %s<br />", geo -> sampleDate);
       g_string_append (outBuffer, "\n\t\t</p>]]></description>");
       if (geo -> sampleDate != NULL) {
         date = DateString (geo -> sampleDate);
         g_string_append (outBuffer, "\n\t\t<TimeStap>");
         g_string_append_printf (outBuffer, "\n\t\t\t<when>%s</when>", date -> str);
         g_string_append (outBuffer, "\n\t\t</TimeStap>");
         g_string_free (date, TRUE);
       }
       g_string_append (outBuffer, "\n\t\t<Point>");
       g_string_append (outBuffer, "\n\t\t\t<extrude>0</extrude>");
       g_string_append (outBuffer, "\n\t\t\t<altitudeMode>absolute</altitudeMode>");
       g_string_append (outBuffer, "\n\t\t\t<coordinates>");
       numStr = g_strdup_printf ("%.6f", geo -> lon);
       CheckRealStr (numStr);
       g_string_append_printf (outBuffer, "%s,", numStr);
       g_free (numStr);
       numStr = g_strdup_printf ("%.6f", geo -> lat);
       CheckRealStr (numStr);
       g_string_append (outBuffer, numStr);
       g_free (numStr);
       g_string_append (outBuffer, "</coordinates>");
       g_string_append (outBuffer, "\n\t\t</Point>\n\t</Placemark>");
     }
     while (navNode != NULL) {
       g_string_append (outBuffer, "\n</Folder>");
       if (g_node_next_sibling (navNode) != NULL) {
         navNode = g_node_next_sibling (navNode);
         break;
       }
       navNode = navNode -> parent;
     }
   }
 }
 g_string_append (outBuffer, "\n\n</Document>\n</kml>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteVRXFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTree;
 PHYLOTREE* tree;
 GPtrArray* recArray;
 FILE* out;
 SCOUNTER i;
 GList* navRec;
 RECONS* actRec;

 iniLen = (85 * GETNUMSPP) + (65 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append (outBuffer, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
 g_string_append (outBuffer, "\n\n<VipResults>");
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recs == NULL) continue;
   for (navRec = tree -> recs; navRec != NULL; navRec = g_list_next (navRec)) {
     g_string_append_printf (outBuffer,"\n\n<Reconstruction>\n\t<treeID>%ld</treeID>", tree -> id);
     recArray = (GPtrArray*) navRec -> data;
     for (i = 0; i < recArray -> len; ++ i) {
       actRec = (RECONS*) g_ptr_array_index (recArray, i);
       if (actRec -> numInher == 0)
         g_string_append_printf (outBuffer, "\n\t<DistroElimination>\n\t\t<nodeID>%ld</nodeID>\n\t</DistroElimination>",
                                 actRec -> node -> id);
       else if (actRec -> numInher < actRec -> numCells)
         g_string_append_printf (outBuffer, "\n\t<PartialElimination>\n\t\t<nodeID>%ld</nodeID>\n\t</PartialElimination>",
                                 actRec -> node -> id);
     }
     g_string_append (outBuffer,"\n</Reconstruction>");
   }
 }
 g_string_append (outBuffer, "\n\n</VipResults>\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteXYDFile (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navTerm;
 GList* navGeo;
 TERMINAL* actTerm;
 GEOREF* geo;
 FILE* out;
 SCOUNTER i;
 gchar* numStr;

 iniLen = (20 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append_printf (outBuffer, "\nspp %ld", GETNUMSPP);
 g_string_append (outBuffer, "\nynegative");
 g_string_append (outBuffer, "\nxydata");
 i = 0;
 for (navTerm = GETSPP; navTerm != NULL; navTerm = g_list_next (navTerm)) {
   actTerm = (TERMINAL*) navTerm -> data;
   if (actTerm -> name != NULL)
     g_string_append_printf (outBuffer, "\n\nsp %ld [%s]", i, actTerm -> name);
   else
     g_string_append_printf (outBuffer, "\n\nsp %ld [species_%ld]", i, i);
   ++ i;
   for (navGeo = actTerm -> firstGeo; navGeo != NULL; navGeo = g_list_next (navGeo)) {
     geo = (GEOREF*) navGeo -> data;
     if (geo -> sp != actTerm -> list) break;
     numStr = g_strdup_printf ("%.6f", geo -> lon);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, "\n%s ", numStr);
     g_free (numStr);
     numStr = g_strdup_printf ("%.6f", geo -> lat);
     CheckRealStr (numStr);
     g_string_append_printf (outBuffer, numStr);
     g_free (numStr);
   }
 }
 g_string_append (outBuffer, "\n\n");
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteTABTable (gchar* fileName) {
 gsize iniLen;
 GString* outBuffer;
 GList* navGeo;
 GEOREF* geo;
 FILE* out;
 gchar* numStr;

 iniLen = (20 * GETNUMGEOREFS);
 outBuffer = g_string_sized_new (iniLen);
 g_string_append_printf (outBuffer,
    "Dataset\tDate collected\tCatalogue No\tIdentifier\tScientific name\tLatitude\tLongitude\tURL\tPrimary reference\n");
 for (navGeo = GETGEOREF; navGeo != NULL; navGeo = g_list_next (navGeo)) {
   geo = (GEOREF*) navGeo -> data;
   if (geo -> collec != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> collec);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> sampleDate != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> sampleDate);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> catalog != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> catalog);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> identifier != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> identifier);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> name != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> name);
   else
     g_string_append (outBuffer, "\t");
   numStr = g_strdup_printf ("%.6f", geo -> lat);
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
   numStr = g_strdup_printf ("%.6f", geo -> lon);
   CheckRealStr (numStr);
   g_string_append_printf (outBuffer, "%s\t", numStr);
   g_free (numStr);
   if (geo -> url != NULL)
     g_string_append_printf (outBuffer, "%s\t", geo -> url);
   else
     g_string_append (outBuffer, "\t");
   if (geo -> reference != NULL)
     g_string_append_printf (outBuffer, "%s\n", geo -> reference);
   else
     g_string_append (outBuffer, "\n");
 }
 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   g_string_free (outBuffer, TRUE);
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
 g_string_free (outBuffer, TRUE);
}

void WriteOutLog (gchar* fileName, GString* outBuffer) {
 FILE* out;

 out = fopen (fileName, "w");
 if (out == NULL) {
   errorMessage = g_strdup_printf ("\nUnable to open %s\n", fileName);
   OnError ();
   return;
 }
 if (fwrite (outBuffer -> str, sizeof (gchar), outBuffer -> len, out) != outBuffer -> len) {
   errorMessage = g_strdup_printf ("\nUnable to save whole data\nFile%s\n", fileName);
   OnError ();
 }
 fclose (out);
}

/* This functions ensure that the out uses points and not commas, This is because my system uses spanish system (i.e. commas) !!" */
void CheckRealStr (gchar* num) {
 SCOUNTER c;

 for (c = 0; ; ++ c) {
   if (num [c] == '.') break;
   if (num [c] == ',') {
     num [c] = '.';
     break;
   }
   if (num [c] == 0) break;
 }
}

GString* DateString (gchar* date) {
 GString* toRet;
 SCOUNTER c;

 toRet = g_string_sized_new (16);
 for (c = 0; ; ++ c) {
   if (date [c] == 0) break;
   if (g_ascii_isspace (date [c])) break;
   g_string_append_c (toRet, date [c]);
 }
 g_string_append_c (toRet, 0);
 return toRet;
}

gchar* CheckExtention (gchar* name, gchar* extention) {
 SCOUNTER c;
 SFLAG noExt;
 gchar* toRet;

 noExt = TRUE;
 for (c = 0; ; ++c) {
   if (name [c] == '\0') {
     noExt = TRUE;
     break;
   }
   if (name [c] == '.') {
     noExt = FALSE;
     continue;
   }
   if ((noExt == FALSE) && (!(g_ascii_isspace (name [c])))) break;
 }
 toRet = name;
 if (noExt == TRUE) {
   toRet = g_strdup_printf ("%s.%s", name, extention);
   g_free (name);
 }
 return toRet;
}
